@@ -1,5 +1,20 @@
+"""
+pyautomation for Python 3.
+Author: iamtony.ca@gmail.com
+Source: https://github.com/changgwak/python-automation
+
+This module is for Automation on Windows os.
+With the pyautomation package, you can control your GUI automatically while simultaneously controlling the mouse and keyboard physically, similar to how selenium automates web browsers.
+Read 'readme.md' for help.
+
+pyautomation is shared under the MIT Licene.
+This means that the code can be freely copied and distributed, and costs nothing to use.
+"""
+
 from . import msuiauto as msauto
 # import msuiauto as msauto
+from .modules import mss
+# import mss
 
 import logging
 import os
@@ -121,6 +136,10 @@ class WinAuto:
         self.config = config
         self.scale_factor = self.get_scale_factor()
 
+        # in order to refine coordinate
+        with mss.mss() as sct:
+            pass
+
     def get_scale_factor(self):
     #     import displayinfo as pydis
     #     scale_factors=pydis.DisplayInfo().get_scale_factor(pydis.DisplayInfo().get_Qapp())
@@ -128,7 +147,7 @@ class WinAuto:
         return None
 
     @log_exceptions(LogLevel.ERROR)
-    def get_info(self, control: Any, depth: int = 0, delimiter: str = "") -> None:
+    def show_info(self, control: Any, depth: int = 0, delimiter: str = "") -> None:
         """Prints information about the control."""
         child_handle = control.NativeWindowHandle
         child_name = control.Name
@@ -137,6 +156,17 @@ class WinAuto:
         child_classname = control.ClassName
         logger.info(f"{delimiter} Depth: {depth}, Window Handle: {child_handle}, name: {child_name}, control_type: {child_controltype}, location: {child_location}, classname: {child_classname}")
 
+    @log_exceptions(LogLevel.ERROR)
+    def get_info(self, control: Any, depth: int = 0, delimiter: str = "") -> Any:
+        """Prints information about the control."""
+        child_handle = control.NativeWindowHandle
+        child_name = control.Name
+        child_controltype = control.LocalizedControlType
+        child_location = control.BoundingRectangle
+        child_classname = control.ClassName
+        return delimiter, depth, child_handle,child_name, child_controltype, child_location, child_classname
+
+
     @log_exceptions(LogLevel.ERROR)
     def get_relative_location(self, parent_control: Any, child_control: Any) -> Tuple[int, int]:
         """Calculates the relative location of the child control to the parent control."""
@@ -154,6 +184,17 @@ class WinAuto:
 
         return relative_center_x, relative_center_y
 
+    @log_exceptions(LogLevel.ERROR)
+    def get_absolute_location(self, child_control: Any) -> Tuple[int, int]:
+        """Calculates the absolute location of the child control[left, top, right, bottom]."""
+        child_rectangle = child_control.BoundingRectangle
+
+        absolute_x = (child_rectangle.left + child_rectangle.right)/2.
+        absolute_y = (child_rectangle.top + child_rectangle.bottom)/2.
+
+        return absolute_x, absolute_y
+
+
     @profile
     @log_exceptions(LogLevel.ERROR)
     def walk_and_find(self, control: Any, depth: int = 0, debug: bool = False) -> Tuple[Optional[Any], Optional[int]]:
@@ -169,6 +210,23 @@ class WinAuto:
                 return result, result_depth
         return None, None
 
+    @profile
+    @log_exceptions(LogLevel.ERROR)
+    def show_walk_and_find_all(self, control: Any, condition: Optional[Callable[[Any], bool]] = None, depth: int = 0) -> List[Tuple[Any, int]]:
+        """Recursively finds all controls that match the condition."""
+        found_controls = []
+        if condition is None:
+            condition = lambda x: True
+
+        if condition(control):
+            found_controls.append((control, depth))
+
+        for child in control.GetChildren():
+            self.show_info(child, depth, "****child")
+            found_controls.extend(self.show_walk_and_find_all(child, condition, depth + 1))
+
+        return found_controls
+
     @profile
     @log_exceptions(LogLevel.ERROR)
     def walk_and_find_all(self, control: Any, condition: Optional[Callable[[Any], bool]] = None, depth: int = 0) -> List[Tuple[Any, int]]:
@@ -186,14 +244,42 @@ class WinAuto:
 
         return found_controls
 
+    @profile
+    @log_exceptions(logging.ERROR)
+    def find_control_with_retry(self, control_class: Any, search_depth: int, name: str, class_name: str, max_retries=None):
+        """Finds a control with retry logic."""
+        retries = 0
+        while max_retries is None or retries < max_retries:
+            try:
+                # Create the control instance using the provided control class
+                control_instance = control_class(searchDepth=search_depth, Name=name, ClassName=class_name)
+                
+                if control_instance is not None:
+                    # print(f"{control_class.__name__} found: {control_instance}")
+                    # logging.info(f"{control_class.__name__} found: {control_instance}")
+                    return control_instance
+                else:
+                    # print(f"{control_class.__name__} not found, retrying...")
+                    logging.warning(f"{control_class.__name__} not found, retrying...")
+            except Exception as e:
+                # print(f"Exception occurred: {e}")
+                logging.error(f"Exception occurred in find_control_with_retry: {e}")
+
+            time.sleep(1)  # Wait for 1 second before retrying
+            retries += 1
+
+        return None
+
+
+
     @log_exceptions(LogLevel.ERROR)
-    def click_relative_location(self, parent_control: Any, x: int, y: int) -> None:
+    def click_relative_location(self, parent_control: Any, x: int, y: int, time: int = 0) -> None:
         """Clicks at a relative location within the parent control."""
         hWnd = parent_control.NativeWindowHandle
         lParam = win32api.MAKELONG(x, y)
         win32gui.PostMessage(hWnd, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)
         win32gui.PostMessage(hWnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, lParam)
-        win32api.Sleep(100)
+        win32api.Sleep(time)
         win32gui.PostMessage(hWnd, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, lParam)
 
     @log_exceptions(LogLevel.ERROR)
@@ -205,28 +291,84 @@ class WinAuto:
         win32gui.PostMessage(hWnd, win32con.WM_KEYUP, win32con.VK_RETURN, 0)
 
     @log_exceptions(LogLevel.ERROR)
-    def click_direct_child(self, child_control: Any) -> None:
+    def click_direct_child(self, child_control: Any, time: int = 0) -> None:
         """Clicks directly on a child control."""
         hwnd = child_control.NativeWindowHandle
         win32gui.PostMessage(hwnd, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)
         win32gui.PostMessage(hwnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, 0)
-        win32api.Sleep(100)
+        win32api.Sleep(time)
         win32gui.PostMessage(hwnd, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, 0)
 
     @log_exceptions(LogLevel.ERROR)
-    def type_text(self, hwnd: int, text: str) -> None:
-        """Types text into a window."""
-        for char in text:
-            if char == "\n":
-                win32gui.PostMessage(hwnd, win32con.WM_KEYDOWN, win32con.VK_RETURN, 0)
-                win32gui.PostMessage(hwnd, win32con.WM_KEYUP, win32con.VK_RETURN, 0)
-            else:
-                win32gui.PostMessage(hwnd, win32con.WM_CHAR, ord(char), 0)
+    def click_direct_hwnd(self, native_window_handle: Any, time: int = 0) -> None:
+        """Clicks directly on a child control."""
+        # if hwnd is hexadecimal, then you need to convert from hexadecimal string to decimal number int. "int(hexadecimal string, 16)".
+        hwnd = native_window_handle
+        win32gui.PostMessage(hwnd, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)
+        win32gui.PostMessage(hwnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, 0)
+        win32api.Sleep(time)
+        win32gui.PostMessage(hwnd, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, 0)
+
+
+    @log_exceptions(LogLevel.ERROR)
+    def send_input(self, hwnd: int, input_data: str | int) -> None:
+        """
+        Sends input to a window. Handles:
+        - Text input (str)
+        - Special keys or single keys (int: virtual key codes)
+        Usage: 
+        - self.send_input(hwnd, "Hello\nWorld")
+        - self.send_input(hwnd, win32con.VK_RETURN) # Enter
+        """
+        if isinstance(input_data, str):
+            for char in input_data:
+                if char == "\n":
+                    self._send_key(hwnd, win32con.VK_RETURN)
+                else:
+                    win32gui.PostMessage(hwnd, win32con.WM_CHAR, ord(char), 0)
+        elif isinstance(input_data, int):
+            self._send_key(hwnd, input_data)
+        else:
+            raise ValueError("Input data must be either a string or an integer (key code).")
+
+    @log_exceptions(LogLevel.ERROR)
+    def send_special_key(self, hwnd: int, key: str) -> None:
+        """
+        Sends special keys like Enter, Arrow keys, etc., to a window using predefined key mappings.
+        Usage:
+        - self.send_special_key(hwnd, "{Enter}")
+        """
+        key_map = {
+            "{Enter}": win32con.VK_RETURN,
+            "{Right Arrow}": win32con.VK_RIGHT,
+            "{Space}": win32con.VK_SPACE,
+            "{Left Arrow}": win32con.VK_LEFT,
+            "{Up Arrow}": win32con.VK_UP,
+            "{Down Arrow}": win32con.VK_DOWN,
+            "{Tab}": win32con.VK_TAB,
+        }
+        
+        key_code = key_map.get(key)
+        if key_code is not None:
+            self._send_key(hwnd, key_code)
+        else:
+            raise ValueError(f"Unsupported special key: {key}")
+
+    @log_exceptions(LogLevel.ERROR)
+    def clear_text(self, hwnd: int) -> None:
+        """Clears the text in a window."""
+        win32gui.SendMessage(hwnd, win32con.WM_SETTEXT, 0, "")
+
+    def _send_key(self, hwnd: int, key_code: int) -> None:
+        """Helper function to send a single key press (keydown and keyup)."""
+        win32gui.PostMessage(hwnd, win32con.WM_KEYDOWN, key_code, 0)
+        win32gui.PostMessage(hwnd, win32con.WM_KEYUP, key_code, 0)
 
     @log_exceptions(LogLevel.ERROR)
     def hotkey_event(self) -> None:
         """Placeholder for handling hotkey events."""
         pass
+    
 
     @log_exceptions(LogLevel.ERROR)
     def get_all_children(self, root: Any) -> None:
@@ -266,16 +408,95 @@ class WinAuto:
         win32api.mouse_event(win32con.MOUSEEVENTF_LEFTDOWN, 0, 0, 0, 0)
         win32api.mouse_event(win32con.MOUSEEVENTF_LEFTUP, 0, 0, 0, 0)
 
-    def _invisible_click(self, x: int, y: int) -> None:
+    def _invisible_click(self, x: int, y: int, time: int = 0) -> None:
         hwnd = win32gui.WindowFromPoint((x, y))
         if hwnd:
             client_coords = win32gui.ScreenToClient(hwnd, (x, y))
             lParam = win32api.MAKELONG(client_coords[0], client_coords[1])
             win32gui.PostMessage(hwnd, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)
             win32gui.PostMessage(hwnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, lParam)
-            win32api.Sleep(100)
+            win32api.Sleep(time)
             win32gui.PostMessage(hwnd, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, lParam)
 
+    @log_exceptions(LogLevel.ERROR)
+    def get_mouse_location(self) -> Tuple[int, int]:
+        """
+        Get the current position of the mouse cursor.
+        :return: Tuple containing the x and y coordinates of the mouse.
+        """
+        try:
+            x, y = win32api.GetCursorPos()
+            logger.info(f"Current mouse location: ({x}, {y})")
+            return x, y
+        except Exception as e:
+            logger.error(f"Error in get_mouse_location: {e}")
+            raise WinAutoError("Failed to get mouse location")
+
+    @log_exceptions(LogLevel.ERROR)
+    def drag_mouse(self, start: List[int], end: List[int], time: int = 1, visible: bool = False) -> None:
+        """
+        Simulate dragging the mouse from one position to another.
+        :param start: Starting x,y coordinate.
+        :param end: Ending x, y coordinate.
+        :time: interval second
+        :param visible: If true, move the mouse cursor visibly. Defaults to False.
+        """
+        try:
+            start_x, start_y = start
+            end_x, end_y = end 
+            duration= time*100
+            
+            # if scale_factor is not None :
+            #     scale_factor_monitor = scale_factor
+            # elif scale_factor is None:
+            #     scale_factor_monitor = self.scale_factor
+
+            # print("scale_factor_monitor: ", scale_factor_monitor)
+            # scaled_x, scaled_y = self._apply_scaling(x, y, scale_factor_monitor)
+            if visible:
+                steps = 100  # Number of steps for smooth dragging
+                step_delay = duration // steps  # Time delay between each step
+                x_step = (end_x - start_x) / steps
+                y_step = (end_y - start_y) / steps
+
+                # Move to the start position and press the left mouse button
+                win32api.SetCursorPos((start_x, start_y))
+                win32api.mouse_event(win32con.MOUSEEVENTF_LEFTDOWN, 0, 0, 0, 0)
+
+                # Smoothly move the mouse in small increments
+                for i in range(steps):
+                    intermediate_x = int(start_x + x_step * i)
+                    intermediate_y = int(start_y + y_step * i)
+                    win32api.SetCursorPos((intermediate_x, intermediate_y))
+                    win32api.Sleep(step_delay)
+
+                # Move to the end position and release the left mouse button
+                win32api.SetCursorPos((end_x, end_y))
+                win32api.mouse_event(win32con.MOUSEEVENTF_LEFTUP, 0, 0, 0, 0)
+            else:
+                # Simulate movement invisibly (background)
+                hwnd = win32gui.WindowFromPoint((start_x, start_y))
+                client_coords = win32gui.ScreenToClient(hwnd, (start_x, start_y))
+                lParam_start = win32api.MAKELONG(client_coords[0], client_coords[1])
+                win32gui.PostMessage(hwnd, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)
+                win32gui.PostMessage(hwnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, lParam_start)
+                win32api.Sleep(duration)                
+
+                hwnd = win32gui.WindowFromPoint((end_x, end_y))
+                client_coords_end = win32gui.ScreenToClient(hwnd, (end_x, end_y))
+                lParam_end = win32api.MAKELONG(client_coords_end[0], client_coords_end[1])
+                
+                win32gui.PostMessage(hwnd, win32con.WM_MOUSEMOVE, win32con.MK_LBUTTON, lParam_end)
+                win32gui.PostMessage(hwnd, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)
+                win32gui.PostMessage(hwnd, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, lParam_end)
+
+            logger.info(f"Mouse dragged from ({start_x}, {start_y}) to ({end_x}, {end_y}) and released")
+
+        except Exception as e:
+            logger.error(f"Error in drag_mouse: {e}")
+            raise WinAutoError("Failed to drag mouse")
+
+
 
 
 
@@ -435,3 +656,6 @@ def test_click_at_invisible(mocker):
     mock_post_message.assert_any_call(12345, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, 100)
 
 # More tests can be added for each method to ensure all edge cases are covered.
+
+
+
